from abc import ABC, abstractmethod
from typing import Union

import pandas as pd


class MetData:
    """
    A class to represent weather data.
    """

    def __init__(self, metadata: dict, data: pd.DataFrame):
        self.metadata = metadata
        self.data = data


class GlmMet(ABC):
    """
    An abstract class for GLM weather data.
    """

    def __init__(
        self,
        location: tuple[float, float],
        date_range: tuple[str, str],
        variables: list[str],
        met_data: Union[None, MetData],
        glm_met_data: Union[None, pd.DataFrame],
    ):
        self.location = location
        self.date_range = date_range
        self.variables = variables
        self.met_data = met_data
        self.glm_met_data = glm_met_data

    @abstractmethod
    def get_variables(self) -> None:
        """Get variables from weather data provider."""
        pass

    @abstractmethod
    def write_met(self) -> None:
        """Save weather data from data provider."""
        pass

    @abstractmethod
    def convert_to_glm(self) -> None:
        """Convert weather data to GLM format."""
        pass

    @abstractmethod
    def write_glm_met(self) -> None:
        """Save weather data in GLM format."""
        pass



"""
Test open-meteo climate API:

test 1 - test initialisation of ClimateChange object.
test 2 - test download of data from open-meteo Climate API.
test 3 - test download of data from open-meteo Climate API with custom variables.
test 4 - test download of data from open-meteo Climate API with custom climate models.
test 5 - test download of data from open-meteo Climate API with custom request parameters.
test 6 - test write of downloaded Climate Change data to file.
test 7 - test conversion of downloaded Climate Change data to GLM format.
test 8 - test write of GLM formatted data to a zipfile.
test 9 - test write of GLM formatted to file. 
"""

import pytest
import tempfile
import os
import pandas as pd
import zipfile

from glm_met.openmeteo import climate_change as climate

def test_climate_init(valid_openmeteo_climate_init):
    """Test initialise climate change object"""
    tmp_climate = climate.ClimateChange(
        location=(valid_openmeteo_climate_init["longitude"], valid_openmeteo_climate_init["latitude"]),
        date_range=valid_openmeteo_climate_init["date_range"],
        variables=valid_openmeteo_climate_init["variables"],
        met_data=None,
        glm_met_data=None
    )

    assert isinstance(tmp_climate, climate.ClimateChange)

def test_climate_get_data(valid_openmeteo_climate_init):
    """Test download of data from open-meteo Climate API"""
    tmp_climate = climate.ClimateChange(
        location=(valid_openmeteo_climate_init["longitude"], valid_openmeteo_climate_init["latitude"]),
        date_range=valid_openmeteo_climate_init["date_range"],
        variables=valid_openmeteo_climate_init["variables"],
        met_data=None,
        glm_met_data=None
    )

    tmp_climate.get_variables(request_settings=None)

    assert isinstance(tmp_climate.met_data.metadata, dict) and isinstance(tmp_climate.met_data.data, pd.core.frame.DataFrame)

def test_climate_get_data_custom_variables(valid_openmeteo_climate_init):
    """Test download of data from open-meteo Climate API with custom variables"""
    tmp_climate = climate.ClimateChange(
        location=(valid_openmeteo_climate_init["longitude"], valid_openmeteo_climate_init["latitude"]),
        date_range=valid_openmeteo_climate_init["date_range"],
        variables=["dewpoint_2m_min", "pressure_msl_mean"],
        met_data=None,
        glm_met_data=None
    )

    tmp_climate.get_variables(request_settings=None)

    assert isinstance(tmp_climate.met_data.metadata, dict) and isinstance(tmp_climate.met_data.data, pd.core.frame.DataFrame)

def test_climate_get_data_custom_models(valid_openmeteo_climate_init):
    """Test download of data from open-meteo Climate API with custom variables"""
    tmp_climate = climate.ClimateChange(
        location=(valid_openmeteo_climate_init["longitude"], valid_openmeteo_climate_init["latitude"]),
        date_range=valid_openmeteo_climate_init["date_range"],
        variables=["dewpoint_2m_min", "pressure_msl_mean"],
        models=["CMCC_CM2_VHR4", "FGOALS_f3_H"],
        met_data=None,
        glm_met_data=None
    )

    tmp_climate.get_variables(request_settings=None)

    meta_units = list(tmp_climate.met_data.metadata["daily_units"].keys())

    if ("pressure_msl_mean_CMCC_CM2_VHR4" in meta_units) and \
            ("pressure_msl_mean_FGOALS_f3_H" in meta_units) and \
                ("dewpoint_2m_min_CMCC_CM2_VHR4" in meta_units) and \
                    ("dewpoint_2m_min_FGOALS_f3_H" in meta_units):
                var_check = 1 


    assert isinstance(tmp_climate.met_data.metadata, dict) and \
          isinstance(tmp_climate.met_data.data, pd.core.frame.DataFrame) and \
          var_check == 1
    
def test_climate_get_data_custom_settings(valid_openmeteo_climate_init):
    """Test download of data from open-meteo Climate API with custom request parameters"""
    tmp_climate = climate.ClimateChange(
        location=(valid_openmeteo_climate_init["longitude"], valid_openmeteo_climate_init["latitude"]),
        date_range=valid_openmeteo_climate_init["date_range"],
        variables=valid_openmeteo_climate_init["variables"],
        met_data=None,
        glm_met_data=None
    )

    settings = {
        "temperature_unit": "fahrenheit"
    }

    tmp_climate.get_variables(request_settings=settings)

    assert isinstance(tmp_climate.met_data.metadata, dict) and isinstance(tmp_climate.met_data.data, pd.core.frame.DataFrame)

def test_climate_write_met(valid_openmeteo_climate_init):
    """Test writing data downloaded from open-meteo Climate API to file"""
    tmp_climate = climate.ClimateChange(
        location=(valid_openmeteo_climate_init["longitude"], valid_openmeteo_climate_init["latitude"]),
        date_range=valid_openmeteo_climate_init["date_range"],
        variables=valid_openmeteo_climate_init["variables"],
        met_data=None,
        glm_met_data=None
    )

    tmp_climate.get_variables(request_settings=None)
    
    with tempfile.TemporaryDirectory() as f:
        tmp_climate.write_met(path=f)
        # check zipfile in tmp dir
        f_files = os.listdir(f)
        with zipfile.ZipFile(os.path.join(f, "met_raw.zip"), "r") as ff:
            f_namelist = ff.namelist()
    
    assert ("met_raw.zip" in f_files) and ("met_raw.csv" in f_namelist)

def test_climate_convert_to_glm(valid_openmeteo_climate_init):
    """Test converting downloaded met data to GLM format"""
    tmp_climate = climate.ClimateChange(
        location=(valid_openmeteo_climate_init["longitude"], valid_openmeteo_climate_init["latitude"]),
        date_range=valid_openmeteo_climate_init["date_range"],
        variables=valid_openmeteo_climate_init["variables"],
        met_data=None,
        glm_met_data=None
    )

    tmp_climate.get_variables(request_settings=None)
    tmp_climate.convert_to_glm()

    # check correct columns generated in DataFrames for all climate models
    matching_cols_all_models = 0
    for m in tmp_climate.models:
        src_glm = tmp_climate.glm_met_data[m]
        src_glm_cols = src_glm.columns
        target_glm_cols = [
            "time",
            "ShortWave",
            "Cloud",
            "AirTemp",
            "RelHum",
            "WindSpeed",
            "Rain"
        ]
        matching_cols = [c for c in src_glm_cols if c  in target_glm_cols]
        if len(matching_cols) == 7:
             matching_cols_all_models += 1
    
    assert (matching_cols_all_models == len(tmp_climate.models)) 

def test_climate_write_glm_zip(valid_openmeteo_climate_init):
    """Test writing GLM formatted data file - zip"""
    tmp_climate = climate.ClimateChange(
        location=(valid_openmeteo_climate_init["longitude"], valid_openmeteo_climate_init["latitude"]),
        date_range=valid_openmeteo_climate_init["date_range"],
        variables=valid_openmeteo_climate_init["variables"],
        met_data=None,
        glm_met_data=None
    )

    tmp_climate.get_variables(request_settings=None)
    tmp_climate.convert_to_glm()
    
    with tempfile.TemporaryDirectory() as f:
        tmp_climate.write_glm_met(f)
        # check zipfile in tmp dir
        f_files = os.listdir(f)
        with zipfile.ZipFile(os.path.join(f, "met_glm.zip"), "r") as ff:
            f_namelist = ff.namelist()
            model_check = True
            for m in tmp_climate.models:
                 if f"met_{m}.csv" not in f_namelist:
                      model_check = False
    
    assert ("met_glm.zip" in f_files) and (model_check)




historical_api_url = "https://archive-api.open-meteo.com/v1/archive"

climate_api_url = "https://climate-api.open-meteo.com/v1/climate"

climate_models = [
    "CMCC_CM2_VHR4",
    "FGOALS_f3_H",
    "HiRAM_SIT_HR",
    "MRI_AGCM3_2_S",
    "EC_Earth3P_HR",
    "MPI_ESM1_2_XR",
    "NICAM16_8S",
]

climate_glm_default = [
    "shortwave_radiation_sum",
    "cloudcover_mean",
    "temperature_2m_mean",
    "temperature_2m_max",
    "temperature_2m_min",
    "relative_humidity_2m_mean",
    "windspeed_10m_mean",
    "precipitation_sum",
    "et0_fao_evapotranspiration_sum",
]

hourly_historical_glm_default = [
    "shortwave_radiation",
    "cloudcover",
    "temperature_2m",
    "relativehumidity_2m",
    "windspeed_10m",
    "precipitation",
]

daily_historical_glm_default = [
    "shortwave_radiation_sum",
    "temperature_2m_mean",
    "temperature_2m_max",
    "temperature_2m_min",
    "windspeed_10m_max",
    "precipitation_sum",
    "et0_fao_evapotranspiration",
]



import json
import os
import tempfile
import zipfile
from typing import Union

import pandas as pd
import requests

from glm_met import glm_met

from . import settings


class MetData(glm_met.MetData):
    def __init__(self, metadata: dict, data: pd.DataFrame):
        """
        Class for meteorological data and its associated metadata.

        Used within instances of the `ClimateChange` class to store
        meteoroligcal data downloaded from the open-meteo
        climate API.

        Attributes
        ----------
        metadata : (dict)
            Metadata associated with the meteorological data.
        data : (pd.DataFrame)
            Meteorological data in a Pandas DataFrame.
        """
        self.metadata = metadata
        self.data = data


class ClimateChange(glm_met.GlmMet):
    """
    Class for retrieving and processing daily climate change data from
    open-meteo's Climate API.

    Examples
    --------
    >>> import os
    >>> import glm_met.openmeteo.climate_change as climate
    >>> clim = climate.ClimateChange(
    ...            location=(116.691155, -34.225812),
    ...            date_range=("2022-01-01", "2032-12-31"),
    ...            met_data=None,
    ...            glm_met_data=None
    ...        )
    >>> clim.get_variables(request_settings=None)
    >>> clim.convert_to_glm()
    >>> clim.write_glm_met(path=os.getcwd())
    """

    # class attribute storing base URL for openmeteo climate API
    climate_api_url = settings.climate_api_url

    def __init__(
        self,
        location: tuple[float, float],
        date_range: tuple[str, str],
        met_data: Union[None, MetData],
        glm_met_data: Union[None, dict],
        models: list[str] = settings.climate_models,
        variables: list[str] = settings.climate_glm_default,
    ):
        """
        Initialize the `ClimateChange` object for retrieving and storing
        meteorological data from open-meteo's Climate API.

        Parameters
        ----------
        location : tuple[float, float]
            Latitude and longitude of the location for the data retrieval.
        date_range : tuple[str, str]
            Start and end dates for the data retrieval
            (in ISO 8601 format, e.g., "YYYY-MM-DD").
        met_data : Union[None, MetData]
            Attribute to store meteorological data downloaded from open-meteo.
        glm_met_data : Union[None, dict]
            Attribute to store meteorological data downloaded from open-meteo
            in GLM format. Dictionary of DataFrames with each DataFrame
            corresponding to a climate model in models.
        models : list[str]
            List of Climate models to download data from.
            Seven models are supported by open-meteo's Climate API:
            https://open-meteo.com/en/docs/climate-api. The default is
            to use all seven models, as specified in `settings.climate_models`.
        variables : list[str]
            List of meteorological variables to retrieve. The default is
            to use the list specified in `settings.climate_glm_default`.
        """
        self.location = location
        self.date_range = date_range
        self.met_data = met_data
        self.glm_met_data = glm_met_data
        self.models = models
        self.variables = variables

    def get_variables(self, request_settings: Union[None, dict]) -> MetData:
        """
        Get variables from the open-meteo Climate API provider and store them
        in the `met_data` attribute.

        Here, windspeed is requested in units of m/s.

        If you need to pass in extra settings to the API, pass these into
        the `get_variables()` method as a dict object. These settings will
        be appended to the API request as query parameters.

        Parameters
        ----------
        request_settings : Union[None, dict]
            Dictionary object of extra settings to pass in as query parameters
            to the open-meteo Climate API. Latitude, longitude, start data,
            end date, daily weather variables, and windspeed units are set from
            the ClimateChange object attributes. Find a list of settings here:
            https://open-meteo.com/en/docs/climate-api
        """

        payload = {
            "longitude": self.location[0],
            "latitude": self.location[1],
            "start_date": self.date_range[0],
            "end_date": self.date_range[1],
            "daily": self.variables,
            "models": self.models,
            "windspeed_unit": "ms",
        }

        # if user has supplied extra query parameters, append them here.
        if request_settings:
            payload = {**payload, **request_settings}

        r = requests.get(self.climate_api_url, params=payload)
        r_json = r.json()
        df = pd.DataFrame(r_json.pop("daily"))
        metadata = r_json

        self.met_data = MetData(metadata=metadata, data=df)

    def write_met(self, path: str) -> None:
        """
        Save meteorological data and its metadata to a zip file.

        Parameters
        ----------
        path : str
            Path to the directory where the zip file should be saved.
        """

        with tempfile.TemporaryDirectory() as tmpdir:
            self.met_data.data.to_csv(
                os.path.join(tmpdir, "met_raw.csv"), index=False
            )

            with open(
                os.path.join(tmpdir, "met_raw_metadata.json"), "w"
            ) as dst:
                json.dump(self.met_data.metadata, dst)

            data_fpath = os.path.join(tmpdir, "met_raw.csv")
            metadata_fpath = os.path.join(tmpdir, "met_raw_metadata.json")

            files = [data_fpath, metadata_fpath]

            with zipfile.ZipFile(
                os.path.join(path, "met_raw.zip"), mode="w"
            ) as z_dst:
                for file in files:
                    z_dst.write(file, arcname=file.split("/")[-1])

    def convert_to_glm(self) -> None:
        """
        Convert meteorological data to GLM format
        and store it in the `glm_met_data` attribute.

        This method expects the `met_data.data` attribute of the `ClimateChange`
        object to have the following variables: `time`, `shortwave_radiation_sum`,
        `cloudcover_mean`, `temperature_2m_mean`, `relative_humidity_2m_mean`,
        `windspeed_10m_mean` and `precipitation_sum`.

        If the data in `met_data.data` has different variables,
        you will need to write your own function to convert the data
        to GLM format.

        `cloudcover` from openmeteo is in percentage. GLM requires it as a proportion
        (0.0 to 1.0).

        `precipitation` is in mm. GLM requires it in m so it is converted.

        Not all meteorological variables required by GLM are available for every
        model. Missing data values will be returned for these cases. Values in the
        `time` column are in `YYYY-MM-DD` format. If `hh:mm:ss` values are
        required, these will need to be added in a subsequent processing step.
        """

        climate_models_glm = {}

        for m in self.models:
            tmp_data = {
                "time": self.met_data.data["time"],
                "ShortWave": self.met_data.data[
                    f"shortwave_radiation_sum_{m}"
                ],
                "Cloud": self.met_data.data[f"cloudcover_mean_{m}"] / 100,
                "AirTemp": self.met_data.data[f"temperature_2m_mean_{m}"],
                "RelHum": self.met_data.data[f"relative_humidity_2m_mean_{m}"],
                "WindSpeed": self.met_data.data[f"windspeed_10m_mean_{m}"],
                "Rain": self.met_data.data[f"precipitation_sum_{m}"] / 1000,
            }

            climate_models_glm[m] = pd.DataFrame(tmp_data)

        self.glm_met_data = climate_models_glm

    def write_glm_met(self, path: str) -> None:
        """
        Save meteorological data in GLM format and its metadata to file.

        Data associated with a different climate model is written to
        a different csv file. The climate models name is included in
        the csv file's name.

        A zip file of csv files is written to `path`.

        Parameters
        ----------
        path : str
            Path to the directory where the zip file should be saved.
        """

        files = []
        with tempfile.TemporaryDirectory() as tmpdir:
            for m in self.models:
                tmp_df = self.glm_met_data[m]
                tmp_df.to_csv(
                    os.path.join(tmpdir, f"met_{m}.csv"), index=False
                )
                files.append(os.path.join(tmpdir, f"met_{m}.csv"))

            with zipfile.ZipFile(
                os.path.join(path, "met_glm.zip"), mode="w"
            ) as z_dst:
                for file in files:
                    z_dst.write(file, arcname=file.split("/")[-1])



import json
import os
import tempfile
import zipfile
from typing import Union

import pandas as pd
import requests

from glm_met import glm_met

from . import settings


class MetData(glm_met.MetData):
    """
    Class for meteorological data and its associated metadata.

    Used within instances of the `Historical` class to store
    meteoroligcal data downloaded from the open-meteo
    historical API.

    Attributes
    ----------
    metadata : (dict)
        Metadata associated with the meteorological data.
    data : (pd.DataFrame)
        Meteorological data in a Pandas DataFrame.
    """

    def __init__(self, metadata: dict, data: pd.DataFrame):
        self.metadata = metadata
        self.data = data


class Historical(glm_met.GlmMet):
    """
    Class for retrieving and processing historical meteorological data from
    open-meteo's historical API.

    Examples
    --------
    >>> import os
    >>> import glm_met.openmeteo.historical as historical
    >>> hist = historical.Historical(
    ...            location=(116.691155, -34.225812),
    ...            date_range=("1970-01-01", "2022-12-31"),
    ...            variables=["temperature_2m", "relativehumidity_2m"],
    ...            met_data=None,
    ...            glm_met_data=None
    ...        )
    >>> hist.get_variables(request_settings=None)
    >>> hist.convert_to_glm()
    >>> hist.write_glm_met(path=os.getcwd(), zip_f=False, fname="met.csv")

    After successful call to `get_variables()`, the `met_data` attribute
    of the `Historical` type object, `hist, is a `MetData` type object.
    This object has two attributes: a metadata dict and a DataFrame of
    met data values.

    >>> metadata = hist.met_data.metadata
    >>> met_data_df = hist.met_data.data
    """

    # class attribute storing base URL for openmeteo historical API
    historical_api_url = settings.historical_api_url

    def __init__(
        self,
        location: tuple[float, float],
        date_range: tuple[str, str],
        met_data: Union[None, MetData],
        glm_met_data: Union[None, pd.DataFrame],
        variables: Union[None, list[str]],
        timezone: str = "auto",
        hourly: bool = True,
    ):
        """Initialises a `Historical` object retrieving and storing
        meteorological data from open-meteo's Historical API.

        Parameters
        ----------
        location : tuple[float, float]
            Latitude and longitude of the location for the data retrieval.
        date_range : tuple[str, str]
            Start and end dates for the data retrieval
            (in ISO 8601 format, e.g., "YYYY-MM-DD").
        met_data : Union[None, MetData]
            Attribute to store meteorological data downloaded from open-meteo.
        glm_met_data : Union[None, pd.DataFrame]
            Attribute to store meteorological data downloaded from open-meteo
            in GLM format.
        variables : Union[None, list[str]]
            List of meteorological variables to retrieve. The default is
            to use the list specified in `settings.hourly_historical_glm_default`
            or `settings.daily_historical_glm_default.
        timezone : str, optional
            Timezone for the data retrieval. Default is "auto".
        hourly : bool, optional
            If True, retrieve hourly data; otherwise, retrieve daily data.
            Default is True.
        """
        self.location = location
        self.date_range = date_range
        self.met_data = met_data
        self.glm_met_data = glm_met_data
        self.timezone = timezone
        self.hourly = hourly
        self.variables = variables

    def get_variables(self, request_settings: Union[None, dict]) -> MetData:
        """
        Get variables from the open-meteo Historical API provider and store them
        in the `met_data` attribute.

        Here, windspeed is requested in units of m/s.

        If you need to pass in extra settings to the API, pass these into
        the `get_variables()` method as a dict object. These settings will
        be appended to the API request as query parameters.

        Parameters
        ----------
        request_settings : Union[None, dict]
            Dictionary object of extra settings to pass in as query parameters to
            the open-meteo Climate API. Latitude, longitude, start data, end date,
            timezone, daily weather variables, and windspeed units are set from the
            `Historical` object attributes. Find a list of settings here:
            https://open-meteo.com/en/docs/historical-weather-api
        """

        # hourly
        if self.hourly:
            if self.variables is None:
                self.variables = settings.hourly_historical_glm_default

            payload = {
                "longitude": self.location[0],
                "latitude": self.location[1],
                "start_date": self.date_range[0],
                "end_date": self.date_range[1],
                "hourly": (",").join(self.variables),
                "timezone": self.timezone,
                "windspeed_unit": "ms",
            }

            # if user has supplied extra query parameters, append them here.
            if request_settings:
                payload = {**payload, **request_settings}

            r = requests.get(self.historical_api_url, params=payload)
            r_json = r.json()
            df = pd.DataFrame(r_json.pop("hourly"))
            metadata = r_json
        # daily
        else:
            if self.variables is None:
                self.variables = settings.daily_historical_glm_default

            payload = {
                "longitude": self.location[0],
                "latitude": self.location[1],
                "start_date": self.date_range[0],
                "end_date": self.date_range[1],
                "daily": (",").join(self.variables),
                "timezone": self.timezone,
                "windspeed_unit": "ms",
            }

            # if user has supplied extra query parameters, append them here.
            if request_settings:
                payload = {**payload, **request_settings}

            r = requests.get(self.historical_api_url, params=payload)
            r_json = r.json()
            df = pd.DataFrame(r_json.pop("daily"))
            metadata = r_json

        self.met_data = MetData(metadata=metadata, data=df)

    def write_met(self, path: str) -> None:
        """
        Save meteorological data and its metadata to a zip file.

        Parameters
        ----------
        path : str
            Path to the directory where the zip file should be saved.
        """

        with tempfile.TemporaryDirectory() as tmpdir:
            self.met_data.data.to_csv(
                os.path.join(tmpdir, "met_raw.csv"), index=False
            )

            with open(
                os.path.join(tmpdir, "met_raw_metadata.json"), "w"
            ) as dst:
                json.dump(self.met_data.metadata, dst)

            data_fpath = os.path.join(tmpdir, "met_raw.csv")
            metadata_fpath = os.path.join(tmpdir, "met_raw_metadata.json")

            files = [data_fpath, metadata_fpath]

            with zipfile.ZipFile(
                os.path.join(path, "met_raw.zip"), mode="w"
            ) as z_dst:
                for file in files:
                    z_dst.write(file, arcname=file.split("/")[-1])

    def convert_to_glm(self) -> None:
        """
        Convert hourly meteorological data to GLM format
        and store it in the `glm_met_data` attribute.

        This function only works with hourly data, this method expects the
        `met_data.data` attribute of the `Historical` object to have the
        following variables: `time`, `shortwave_radiation`, `cloudcover`,
        `temperature_2m`, `relativehumidity_2m`, `windspeed_10m` and
        `precipitation`.

        If the data in `met_data.data` has different
        variables, you will need to write your own function to convert the data
        to GLM format.

        `cloudcover` from openmeteo is in percentage units. GLM requires it as
        a proportion (0.0 to 1.0).

        `precipitation` is in mm. GLM requires it in m / day even if 
        using hourly data, so it is converted.
        """

        if self.hourly:
            df_glm = pd.DataFrame(
                # glm requires time to have "hh:mm:ss" - append :00 to date-time string
                data={
                    "time": self.met_data.data["time"]
                    .str.replace("T", " ")
                    .astype(str)
                    + ":00",
                    "ShortWave": self.met_data.data["shortwave_radiation"],
                    "Cloud": self.met_data.data["cloudcover"] / 100,
                    "AirTemp": self.met_data.data["temperature_2m"],
                    "RelHum": self.met_data.data["relativehumidity_2m"],
                    "WindSpeed": self.met_data.data["windspeed_10m"],
                    "Rain": (self.met_data.data["precipitation"] / 1000) * 24,
                }
            )

            self.glm_met_data = df_glm

    def write_glm_met(self, path: str, zip_f: bool, fname: str) -> None:
        """
        Save meteorological data in GLM format and its metadata to file.

        Only works after a call to `convert_to_glm()` and with hourly
        data.

        Parameters
        ----------
        path : str
            Path to the directory where the zip file should be saved.
        zip_f : bool
            Whether to save output to a zip file.
        fname : str
            If zip_f is `True`, filename for GLM meteorological data csv.
        """

        if self.hourly:
            glm_metadata = self.met_data.metadata.pop("hourly_units")

            if zip_f:
                with tempfile.TemporaryDirectory() as tmpdir:
                    self.glm_met_data.to_csv(
                        os.path.join(tmpdir, "met.csv"), index=False
                    )

                    with open(
                        os.path.join(tmpdir, "met_glm_metadata.json"), "w"
                    ) as dst:
                        json.dump(glm_metadata, dst)

                    data_fpath = os.path.join(tmpdir, "met.csv")
                    metadata_fpath = os.path.join(
                        tmpdir, "met_glm_metadata.json"
                    )

                    files = [data_fpath, metadata_fpath]

                    with zipfile.ZipFile(
                        os.path.join(path, "met_glm.zip"), mode="w"
                    ) as z_dst:
                        for file in files:
                            z_dst.write(file, arcname=file.split("/")[-1])
            else:
                self.glm_met_data.to_csv(
                    os.path.join(path, fname), index=False
                )



"""
Test open-meteo historical API:

test 1 - test initialisation of Historical object.
test 2 - test download of hourly data from open-meteo Historical API.
test 3 - test download of data from open meteo Historical API 
    with custom request settings passed in as query parameters
    and changing timezone.
test 4 - test download of daily data from open-meteo Historical API.
test 5 - test download of daily data from open-meteo Historical API
    with user selected weather variables.
test 6 - test write raw downloaded meteorological data to file.
test 7 - test convert data from open-meteo Historical API to GLM format.
test 8 - test writing GLM formatted data to a zipfile.
test 9 - test writing GLM formatted data to file.
"""

import pytest
import tempfile
import os
import pandas as pd
import zipfile

from glm_met.openmeteo import historical

def test_historical_init(valid_openmeteo_historical_init):
    """Test initialise histoical object"""
    tmp_historical = historical.Historical(
        location=(valid_openmeteo_historical_init["longitude"], valid_openmeteo_historical_init["latitude"]),
        date_range=valid_openmeteo_historical_init["date_range"],
        variables=valid_openmeteo_historical_init["variables"],
        met_data=None,
        glm_met_data=None
    )

    assert isinstance(tmp_historical, historical.Historical)

def test_historical_get_data(valid_openmeteo_historical_init):
    """Test download of data from open-meteo Historical API"""
    tmp_historical = historical.Historical(
        location=(valid_openmeteo_historical_init["longitude"], valid_openmeteo_historical_init["latitude"]),
        date_range=valid_openmeteo_historical_init["date_range"],
        variables=valid_openmeteo_historical_init["variables"],
        met_data=None,
        glm_met_data=None
    )

    tmp_historical.get_variables(request_settings=None)

    assert isinstance(tmp_historical.met_data.metadata, dict) and isinstance(tmp_historical.met_data.data, pd.core.frame.DataFrame)

def test_historical_get_data_custom_settings(valid_openmeteo_historical_init):
    """Test download of data from open-meteo Historical API
    
    Pass in custom settings and set timezone.

    """
    tmp_historical = historical.Historical(
        location=(valid_openmeteo_historical_init["longitude"], valid_openmeteo_historical_init["latitude"]),
        date_range=valid_openmeteo_historical_init["date_range"],
        variables=valid_openmeteo_historical_init["variables"],
        met_data=None,
        glm_met_data=None,
        timezone="Australia/Sydney"
    )

    settings = {
        "temperature_unit": "fahrenheit",
        "precipitation_unit": "inch" 
    }

    tmp_historical.get_variables(request_settings=settings)

    assert isinstance(tmp_historical.met_data.metadata, dict) and isinstance(tmp_historical.met_data.data, pd.core.frame.DataFrame)

def test_historical_get_data_daily(valid_openmeteo_historical_init):
    """Test download of daily data from open-meteo Historical API

    Get daily data using default daily variables from settings.py. 
    """
    tmp_historical = historical.Historical(
        location=(valid_openmeteo_historical_init["longitude"], valid_openmeteo_historical_init["latitude"]),
        date_range=valid_openmeteo_historical_init["date_range"],
        variables=None,
        met_data=None,
        glm_met_data=None,
        timezone="Australia/Sydney",
        hourly=False
    )

    tmp_historical.get_variables(request_settings=None)

    target_cols = [
    "shortwave_radiation_sum",
    "temperature_2m_mean",
    "temperature_2m_max",
    "temperature_2m_min",
    "windspeed_10m_max",
    "precipitation_sum",
    "et0_fao_evapotranspiration"
    ]   

    src_cols = tmp_historical.met_data.data.columns

    matching_cols = [c for c in src_cols if c in target_cols]


    assert isinstance(tmp_historical.met_data.metadata, dict) \
        and isinstance(tmp_historical.met_data.data, pd.core.frame.DataFrame) \
        and (len(matching_cols) == 7)
    
def test_historical_get_data_daily_custom_variables(valid_openmeteo_historical_init):
    """Test download of daily data from open-meteo Historical API

    Get daily data using default daily variables from settings.py. 
    """
    tmp_historical = historical.Historical(
        location=(valid_openmeteo_historical_init["longitude"], valid_openmeteo_historical_init["latitude"]),
        date_range=valid_openmeteo_historical_init["date_range"],
        variables=["shortwave_radiation_sum"],
        met_data=None,
        glm_met_data=None,
        timezone="Australia/Sydney",
        hourly=False
    )

    tmp_historical.get_variables(request_settings=None)

    target_cols = [
    "shortwave_radiation_sum"
    ]   

    src_cols = tmp_historical.met_data.data.columns

    matching_cols = [c for c in src_cols if c in target_cols]


    assert isinstance(tmp_historical.met_data.metadata, dict) \
        and isinstance(tmp_historical.met_data.data, pd.core.frame.DataFrame) \
        and (len(matching_cols) == 1)
        
def test_historical_write_met(valid_openmeteo_historical_init):
    """Test writing data downloaded from open-meteo Historical API to file"""
    tmp_historical = historical.Historical(
        location=(valid_openmeteo_historical_init["longitude"], valid_openmeteo_historical_init["latitude"]),
        date_range=valid_openmeteo_historical_init["date_range"],
        variables=valid_openmeteo_historical_init["variables"],
        met_data=None,
        glm_met_data=None
    )

    tmp_historical.get_variables(request_settings=None)
    
    with tempfile.TemporaryDirectory() as f:
        tmp_historical.write_met(path=f)
        # check zipfile in tmp dir
        f_files = os.listdir(f)
        with zipfile.ZipFile(os.path.join(f, "met_raw.zip"), "r") as ff:
            f_namelist = ff.namelist()
    
    assert ("met_raw.zip" in f_files) and ("met_raw.csv" in f_namelist)

def test_historical_convert_to_glm(valid_openmeteo_historical_init):
    """Test converting downloaded met data to GLM format"""
    tmp_historical = historical.Historical(
        location=(valid_openmeteo_historical_init["longitude"], valid_openmeteo_historical_init["latitude"]),
        date_range=valid_openmeteo_historical_init["date_range"],
        variables=valid_openmeteo_historical_init["variables"],
        met_data=None,
        glm_met_data=None
    )

    tmp_historical.get_variables(request_settings=None)
    tmp_historical.convert_to_glm()
    src_glm_cols = tmp_historical.glm_met_data.columns
    target_glm_cols = [
        "time",
        "ShortWave",
        "Cloud",
        "AirTemp",
        "RelHum",
        "WindSpeed",
        "Rain"
    ]
    matching_cols = [c for c in src_glm_cols if c  in target_glm_cols]
    
    assert (len(matching_cols) == 7) and (tmp_historical.met_data.data.shape[0] == tmp_historical.glm_met_data.shape[0])

def test_historical_write_glm_zip(valid_openmeteo_historical_init):
    """Test writing GLM formatted data file - zip"""
    tmp_historical = historical.Historical(
        location=(valid_openmeteo_historical_init["longitude"], valid_openmeteo_historical_init["latitude"]),
        date_range=valid_openmeteo_historical_init["date_range"],
        variables=valid_openmeteo_historical_init["variables"],
        met_data=None,
        glm_met_data=None
    )

    tmp_historical.get_variables(request_settings=None)
    tmp_historical.convert_to_glm()
    
    with tempfile.TemporaryDirectory() as f:
        tmp_historical.write_glm_met(f, True, None)
        # check zipfile in tmp dir
        f_files = os.listdir(f)
        with zipfile.ZipFile(os.path.join(f, "met_glm.zip"), "r") as ff:
            f_namelist = ff.namelist()
    
    assert ("met_glm.zip" in f_files) and ("met.csv" in f_namelist)

def test_historical_write_glm(valid_openmeteo_historical_init):
    """Test writing GLM formatted data file"""
    tmp_historical = historical.Historical(
        location=(valid_openmeteo_historical_init["longitude"], valid_openmeteo_historical_init["latitude"]),
        date_range=valid_openmeteo_historical_init["date_range"],
        variables=valid_openmeteo_historical_init["variables"],
        met_data=None,
        glm_met_data=None
    )

    tmp_historical.get_variables(request_settings=None)
    tmp_historical.convert_to_glm()
    
    with tempfile.TemporaryDirectory() as f:
        tmp_historical.write_glm_met(f, False, "test.csv")
        f_files = os.listdir(f)
    
    assert ("test.csv" in f_files)